"""
svgen - A module for XML-style attribute interfaces.
"""

# built-in
from abc import ABC, abstractmethod


class Attribute(ABC):
    """An interface for an XML-style attribute."""

    def __eq__(self, other: object) -> bool:
        """Determine if two attributes are the same."""

        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key == other.key and self.value == other.value

    @property
    def key(self) -> str:
        """Get the string key for this attribute."""

        if hasattr(self, "name"):
            return getattr(self, "name")

        key = type(self).__name__
        return key[0].lower() + key[1:]

    @property
    @abstractmethod
    def value(self) -> str:
        """Get the string value for this attribute."""

    def encode(self, quote: str = '"') -> str:
        """Convert this attribute to a string."""

        return f"{self.key}={quote}{self.value}{quote}"

    @staticmethod
    @abstractmethod
    def decode(key: str, value: str) -> "Attribute":
        """Create this attribute from a string."""


class SimpleAttribute(Attribute):
    """A simple, concrete implementation for attribute."""

    def __init__(self, name: str, value: str) -> None:
        """
        Construct a simple attribute, one with just a key (name) and value.
        """

        self.name = name
        self._value = value

    @property
    def value(self) -> str:
        """Get the string value for this attribute."""

        return self._value

    @staticmethod
    def decode(key: str, value: str) -> Attribute:
        """Create this attribute from a string."""

        return SimpleAttribute(key, value)


XMLNS = SimpleAttribute("xmlns", "http://www.w3.org/2000/svg")



"""
svgen - A module for the 'circle' element.
"""

# built-in
from typing import List

# internal
from svgen.attribute import Attribute
from svgen.attribute.viewbox import ViewBox
from svgen.cartesian import UNITY, Point, Translation, to_center
from svgen.cartesian.circle import Circle as CartCircle
from svgen.element import Element


class Circle(Element):
    """A class for circle elements."""

    def __init__(self, circle: CartCircle) -> None:
        """Construct a new circle element."""

        self.raw = circle
        assert self.raw.center.center

        attrs: List[Attribute] = []
        attrs.append(self.raw.radius_attr)
        attrs += [*self.raw.center_attrs]

        super().__init__(attributes=attrs)

    def translate(self, move: Translation) -> "Circle":
        """Move this circle by a given translation."""
        return Circle(self.raw.translate(move))

    def scale(self, scale: float = UNITY) -> "Circle":
        """Scale this circle's radius."""
        return Circle(self.raw.scale(scale))

    @property
    def radius(self) -> float:
        """Get this circle's radius."""
        return self.raw.radius

    @property
    def center(self) -> Point:
        """Get this circle's center."""
        return self.raw.center

    def __eq__(self, other: object) -> bool:
        """Determine if this circle is equivalent to another."""

        if not isinstance(other, Circle):
            return NotImplemented
        return self.raw == other.raw


def centered(box: ViewBox, radius_scale: float = UNITY) -> Circle:
    """
    From a viewBox, create a circle that is centered with an appropriately
    scaled radius.
    """

    radius = float(min(box.data.width, box.data.height)) / 2.0
    return Circle(
        CartCircle(radius, to_center(box.center)).scale(radius_scale)
    )



"""
svgen - This package's command-line entry-point application.
"""

# built-in
import argparse
from json import load
from pathlib import Path
from typing import cast

# internal
from svgen import PKG_NAME
from svgen.attribute.viewbox import ViewBox
from svgen.config import Config
from svgen.element.svg import Svg
from svgen.script import invoke_script


def entry(args: argparse.Namespace) -> int:
    """Execute the requested task."""

    # Load the top-level configuration.
    try:
        with args.config.open(encoding="utf-8") as config_fd:
            config = Config(load(config_fd))

    # Use a default configuration, if we can't find one,
    except FileNotFoundError:
        config = Config()

    config.set_if_not("height", args.height)
    config.set_if_not("width", args.width)
    config.set_if_not("scripts", [])

    # Compose the document, via the external script.
    doc = Svg(ViewBox.from_dict(cast(dict, config)))
    for script in args.scripts + [Path(x) for x in config["scripts"]]:
        invoke_script(script, doc, config)

    # Write the composed document to the output file.
    with args.output.open("w", encoding="utf-8") as output_fd:
        doc.encode(output_fd)

    return 0


def add_app_args(parser: argparse.ArgumentParser) -> None:
    """Add application-specific arguments to the command-line parser."""

    parser.add_argument(
        "-c",
        "--config",
        type=Path,
        default=Path(f"{PKG_NAME}.json"),
        help="top-level configuration to load (default: '%(default)s')",
    )

    parser.add_argument(
        "--height",
        type=int,
        default=100,
        help=(
            "height of the document, if not specified by "
            "configuration (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--width",
        type=int,
        default=100,
        help=(
            "width of the document, if not specified by "
            "configuration (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=Path(f"{PKG_NAME}.svg"),
        help="file to write SVG output (default: '%(default)s')",
    )

    parser.add_argument(
        "scripts",
        type=Path,
        nargs="*",
        help="scripts to run for composing the SVG document (in order)",
    )



"""
svgen - A module for the 'svg' element.
"""

# built-in
from typing import List

# internal
from svgen.attribute import XMLNS, Attribute
from svgen.attribute.viewbox import ViewBox
from svgen.element import Element


class Svg(Element):
    """A class for svg elements."""

    def __init__(self, viewbox: ViewBox, document: bool = True) -> None:
        """Construct a new svg element (or document)."""

        self.viewbox = viewbox
        attrs: List[Attribute] = [self.viewbox]
        if document:
            attrs.append(XMLNS)
        super().__init__(attributes=attrs)



"""
svgen - A module for Cartesian-coordinate interfaces.
"""

# built-in
from math import isclose, sqrt
from typing import NamedTuple

# internal
from svgen.attribute import SimpleAttribute


class PointAttrs(NamedTuple):
    """A grouping for 'x' and 'y' attributes."""

    x: SimpleAttribute
    y: SimpleAttribute

    @property
    def x_val(self) -> float:
        """Get the 'x' coordinate for these attributes."""

        return float(self.x.value)

    @property
    def y_val(self) -> float:
        """Get the 'y' coordinate for these attributes."""

        return float(self.y.value)


class Translation(NamedTuple):
    """A definition of a translation in a Cartesian coordinate system."""

    dx: float = float()
    dy: float = float()


class Point(NamedTuple):
    """A definition of a point in a Cartesian coordinate system."""

    x: float = float()
    y: float = float()
    center: bool = False

    def translate(self, move: Translation) -> "Point":
        """Move a point by a given translation."""

        return Point(self.x + move.dx, self.y + move.dy, self.center)

    @property
    def x_attr(self) -> SimpleAttribute:
        """Get the 'x' attribute for this point."""

        attr = "x"
        if self.center:
            attr = "c" + attr
        return SimpleAttribute(attr, str(self.x))

    @property
    def y_attr(self) -> SimpleAttribute:
        """Get the 'y' attribute for this point."""

        attr = "y"
        if self.center:
            attr = "c" + attr
        return SimpleAttribute(attr, str(self.y))

    @property
    def attrs(self) -> PointAttrs:
        """Get the 'x' and 'y' attributes for this point."""

        return PointAttrs(self.x_attr, self.y_attr)

    def __eq__(self, other: object) -> bool:
        """Determine if two points are the same."""

        if not isinstance(other, Point):
            return NotImplemented
        return isclose(self.x, other.x) and isclose(self.y, other.y)

    def distance(self, point: "Point") -> float:
        """Compute the distance from this point to another."""

        run = point.x - self.x
        rise = point.y - self.y
        return sqrt(run * run + rise * rise)


def to_center(point: Point) -> Point:
    """Convert an existing point to one that is definitely a 'center' type."""

    return Point(point.x, point.y, True)


DEFAULT = Point()
DEFAULT_CENTER = to_center(DEFAULT)
UNITY: float = 1.0


def distance(point_a: Point, point_b: Point = DEFAULT) -> float:
    """Compute the distance between two points."""

    return point_a.distance(point_b)



"""
svgen - Common interfaces and assets for SVG elements.
"""

# built-in
from io import StringIO
import os
from typing import Dict, List, TextIO

# internal
from svgen.attribute import Attribute

INDENT: int = 2


class Element:
    """A basic interface for an SVG element."""

    def __init__(
        self,
        name: str = None,
        content: str = "",
        attributes: List[Attribute] = None,
        children: List["Element"] = None,
    ) -> None:
        """Construct a new SVG element."""

        if name is None:
            name = type(self).__name__
            name = name[0].lower() + name[1:]
        if attributes is None:
            attributes = []
        if children is None:
            children = []

        self.name = name
        self.content = content

        self.attributes: Dict[str, Attribute] = {}
        for attr in attributes:
            self.add_attribute(attr)

        self.children: List[Element] = children

    def add_attribute(self, attr: Attribute) -> "Element":
        """Add an attribute to this element."""
        assert attr.key not in self.attributes
        self.attributes[attr.key] = attr
        return self

    def closing(self, indent: int = 0) -> str:
        """Create a string to close this element."""

        if not self.content and not self.children:
            return " />"

        indent_str = " " * (indent * INDENT)
        return f"{indent_str}</{self.name}>"

    def encode(
        self,
        output: TextIO,
        quote: str = '"',
        indent: int = 0,
        newlines: bool = True,
    ) -> None:
        """Encode this element to a string stream."""

        # Indent will matter if we want lines.
        indent_str = " " * (indent * INDENT)
        if newlines:
            output.write(indent_str)

        attrs = " ".join(x.encode(quote) for x in self.attributes.values())
        output.write(f"<{self.name}")
        if attrs:
            output.write(f" {attrs}")

        if self.content or self.children:
            # Close the opening tag.
            output.write(">")
            if newlines:
                output.write(os.linesep)

            # Write content, if any.
            if newlines:
                output.write(indent_str + " " * INDENT)
            if self.content:
                output.write(self.content)
                if newlines:
                    output.write(os.linesep)

            # Write children.
            for child in self.children:
                child.encode(output, quote, indent + 1, newlines)

        # Close this element tag.
        output.write(self.closing(indent))
        if newlines:
            output.write(os.linesep)

    def encode_str(
        self,
        quote: str = '"',
        indent: int = 0,
        newlines: bool = True,
    ) -> str:
        """Encode this element to a string."""

        with StringIO() as output:
            self.encode(output, quote, indent, newlines)
            return output.getvalue()



"""
svgen - A module for interacting with rectangular entities.
"""

# built-in
from math import isclose
from typing import NamedTuple

# internal
from svgen.attribute import SimpleAttribute
from svgen.cartesian import DEFAULT, UNITY, Point, Translation


class DimensionAttrs(NamedTuple):
    """A grouping for 'width' and 'height' attributes."""

    width: SimpleAttribute
    height: SimpleAttribute

    @property
    def width_val(self) -> float:
        """Get this 'width' value for these dimensions."""

        return float(self.width.value)

    @property
    def height_val(self) -> float:
        """Get the 'height' value for these dimensions."""

        return float(self.height.value)


class Dimensions(NamedTuple):
    """A set of dimensions describing a rectangular entity."""

    width: float
    height: float

    def __eq__(self, other: object) -> bool:
        """Determine if these dimensions are equivalent to another set."""

        if not isinstance(other, Dimensions):
            return NotImplemented
        return isclose(self.width, other.width) and isclose(
            self.height, other.height
        )

    @property
    def width_attr(self) -> SimpleAttribute:
        """Get the 'width' attribute for these dimensions."""

        return SimpleAttribute("width", str(self.width))

    @property
    def height_attr(self) -> SimpleAttribute:
        """Get the 'height' attribute for these dimensions."""

        return SimpleAttribute("height", str(self.height))

    @property
    def attrs(self) -> DimensionAttrs:
        """Get the 'width' and 'height' attributes for these dimensions."""

        return DimensionAttrs(self.width_attr, self.height_attr)

    @property
    def square(self) -> bool:
        """Determine if these dimensions are square."""

        return isclose(self.width, self.height)

    def scale(
        self, width_scale: float = UNITY, height_scale: float = UNITY
    ) -> "Dimensions":
        """Scale these dimensions."""

        return Dimensions(self.width * width_scale, self.height * height_scale)

    def to_square(self, scale: float = UNITY) -> "Dimensions":
        """Convert these dimensions to a scaled square."""

        length = min(self.width, self.height)
        return Dimensions(length, length).scale(scale, scale)


class Rectangle(NamedTuple):
    """A definition of a rectangle."""

    dimensions: Dimensions
    location: Point = DEFAULT

    def __eq__(self, other: object) -> bool:
        """Determine if this rectangle is equivalent to another."""

        if not isinstance(other, Rectangle):
            return NotImplemented
        return (
            self.dimensions == other.dimensions
            and self.location == other.location
        )

    @property
    def square(self) -> bool:
        """Determine if this rectangle is square."""

        return self.dimensions.square

    def translate(self, move: Translation) -> "Rectangle":
        """Move a rectangle by a given translation."""

        return Rectangle(self.dimensions, self.location.translate(move))

    def scale(
        self, width_scale: float = UNITY, height_scale: float = UNITY
    ) -> "Rectangle":
        """Scale this rectangle's width and height."""

        return Rectangle(
            self.dimensions.scale(width_scale, height_scale), self.location
        )

    def scale_whole(self, scalar: float = UNITY) -> "Rectangle":
        """Scale width and height of this rectangle equally."""

        return self.scale(scalar, scalar)

    def to_square(self, scale: float = UNITY) -> "Rectangle":
        """Convert this rectangle to a square."""

        return Rectangle(self.dimensions.to_square(scale), self.location)



"""
svgen - A class for managing command-line configuration data.
"""

# built-in
from collections import UserDict


class Config(UserDict):
    """A dictionary that allows access to data only once via each key."""

    def __getitem__(self, path):
        """Consume keys after their data is retreived."""
        result = self.data[path]
        del self.data[path]
        return result

    def set_if_not(self, key, value):
        """Set this value if a value is not already set."""
        if key not in self.data:
            self.data[key] = value



"""
svgen - A module for working with external scripts.
"""

# built-in
import importlib.machinery
import importlib.util
from pathlib import Path

# internal
from svgen.config import Config
from svgen.element.svg import Svg


def invoke_script(script: Path, svg: Svg, config: Config) -> None:
    """Invoke an external script's 'compose' function."""

    loader = importlib.machinery.SourceFileLoader("script", str(script))
    spec = importlib.util.spec_from_loader("script", loader)
    assert spec is not None
    script_module = importlib.util.module_from_spec(spec)
    loader.exec_module(script_module)

    # Compose the document.
    svg.children.extend(script_module.compose(svg.viewbox, config))



"""
svgen - A module for the 'style' attribute.
"""

# built-in
from typing import List, NamedTuple

# internal
from svgen.attribute import Attribute
from svgen.color import Color


class CssProperty(NamedTuple):
    """A class for interacting with css properties."""

    key: str
    value: str

    def __eq__(self, other: object) -> bool:
        """Determine if two css properties are the same."""

        if not isinstance(other, CssProperty):
            return NotImplemented
        return self.key == other.key and self.value == other.value

    @staticmethod
    def color(color: Color, prop: str = "fill") -> "CssProperty":
        """Get a CSS property for a color."""
        return CssProperty(prop, str(color))

    @staticmethod
    def decode(value: str) -> List["CssProperty"]:
        """Create css properties from a string."""

        result: List[CssProperty] = []

        value = value.strip()
        properties = [x.strip() for x in value.split(";")]
        for prop in properties:
            if prop:
                tokens = prop.split(":")
                assert len(tokens) == 2
                key = tokens[0].strip()
                val = tokens[1].strip()
                result.append(CssProperty(key, val))

        return result

    @property
    def encoded(self) -> str:
        """Convert this css property to a string."""
        return f"{self.key}: {self.value}"

    @staticmethod
    def encode(properties: List["CssProperty"]) -> str:
        """Encode a list of css properties to a string."""
        return "; ".join(x.encoded for x in properties)


class Style(Attribute):
    """An interface for style attributes."""

    def __init__(self, properties: List[CssProperty] = None) -> None:
        """Construct a new style attribute."""

        if properties is None:
            properties = []
        self.properties: List[CssProperty] = properties

    def __eq__(self, other: object) -> bool:
        """Determine if two style attributes are the same."""

        if not isinstance(other, Style):
            return NotImplemented

        result = self.key == other.key
        result = result and len(self.properties) == len(other.properties)
        if result:
            props = sorted(self.properties)
            other_props = sorted(other.properties)
            for idx, prop in enumerate(props):
                if not result:
                    break
                result = result and prop == other_props[idx]

        return result

    def add_color(self, color: Color, prop: str = "fill") -> None:
        """Set a style property to a color."""
        self.properties.append(CssProperty.color(color, prop))

    @property
    def value(self) -> str:
        """Get the string value for this attribute."""
        return CssProperty.encode(self.properties)

    @staticmethod
    def decode(key: str, value: str) -> Attribute:
        """Create this attribute from a string."""

        assert key == "style"
        return Style(CssProperty.decode(value))



"""
svgen - A module for the 'rect' element.
"""

# built-in
from math import isclose
from typing import List

# internal
from svgen.attribute import Attribute, SimpleAttribute
from svgen.attribute.viewbox import ViewBox
from svgen.cartesian import UNITY, Translation
from svgen.cartesian.rectangle import Rectangle
from svgen.element import Element


class Rect(Element):
    """A class for rect elements."""

    def __init__(
        self,
        rect: Rectangle,
        radius_x: float = float(),
        radius_y: float = float(),
    ) -> None:
        """Construct a new rect element."""

        self.rect = rect
        self.location = self.rect.location
        self.dimensions = self.rect.dimensions
        self.radius_x = radius_x
        self.radius_y = radius_y

        attrs: List[Attribute] = []
        attrs += [*self.location.attrs]
        attrs += [*self.dimensions.attrs]

        if self.radius_x != float():
            attrs.append(SimpleAttribute("rx", str(self.radius_x)))
        if self.radius_y != float():
            attrs.append(SimpleAttribute("ry", str(self.radius_y)))

        super().__init__(attributes=attrs)

    @property
    def square(self) -> bool:
        """Determine if this rectangle is square."""

        return self.rect.square

    def to_square(self, scale: float = UNITY) -> "Rect":
        """Convert this rectangle to a square."""

        return Rect(self.rect.to_square(scale), self.radius_x, self.radius_y)

    def translate(self, move: Translation) -> "Rect":
        """Move this rectangle by a given translation."""

        return Rect(self.rect.translate(move), self.radius_x, self.radius_y)

    def scale(
        self, width_scale: float = UNITY, height_scale: float = UNITY
    ) -> "Rect":
        """Scale this rectangle's width and height."""

        return Rect(
            self.rect.scale(width_scale, height_scale),
            self.radius_x,
            self.radius_y,
        )

    def scale_whole(self, scalar: float = UNITY) -> "Rect":
        """Scale width and height of this rectangle equally."""

        return self.scale(scalar, scalar)

    def __eq__(self, other: object) -> bool:
        """Determine if this rectangle is equivalent to another."""

        if not isinstance(other, Rect):
            return NotImplemented
        return (
            self.rect == other.rect
            and isclose(self.radius_x, other.radius_x)
            and isclose(self.radius_y, other.radius_y)
        )


def centered(
    box: ViewBox, width_scale: float = UNITY, height_scale: float = UNITY
) -> Rect:
    """From a viewBox, created a centered-and-scaled rectangle."""

    dimensions = box.dimensions.scale(width_scale, height_scale)
    delta_x = (box.dimensions.width - dimensions.width) / 2.0
    delta_y = (box.dimensions.height - dimensions.height) / 2.0
    Translation(delta_x, delta_y)
    return Rect(
        Rectangle(
            dimensions, box.origin.translate(Translation(delta_x, delta_y))
        )
    )



# =====================================
# generator=datazen
# version=1.12.1
# hash=590d398349b626cdf55efaac20ebb3f7
# =====================================

"""
vmklib - Useful defaults and other package metadata.
"""

DESCRIPTION = "Simplify project workflows by standardizing use of GNU Make."
PKG_NAME = "vmklib"
VERSION = "0.7.6"



# =====================================
# generator=datazen
# version=1.13.3
# hash=025fef3d48d942de1e95d9b4ef27a4ac
# =====================================

"""
svgen - Useful defaults and other package metadata.
"""

DESCRIPTION = "A tool for working with scalable vector graphics."
PKG_NAME = "svgen"
VERSION = "0.2.1"



"""
vmklib - This package's command-line entry-point application.
"""

# built-in
import argparse
from contextlib import contextmanager
import logging
import os
from pathlib import Path
import subprocess
import tempfile
from typing import Iterator

# third-party
import pkg_resources

LOG = logging.getLogger(__name__)
DEFAULT_FILE = Path("Makefile")


def get_resource(resource_name: str) -> Path:
    """Locate the path to a package resource."""

    resource_path = os.path.join("data", resource_name)

    locations = [
        pkg_resources.resource_filename(__name__, resource_path),
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), resource_path
        ),
    ]

    resource = ""
    for location in locations:
        if os.path.isfile(location):
            resource = location
            break

    # ensure that the resource can actually be found
    assert resource, f"Couldn't load resource '{resource_name}'!"
    return Path(resource)


@contextmanager
def build_makefile(
    user_file: Path, directory: Path, project_name: str = None
) -> Iterator[str]:
    """Build a temporary makefile and return the path."""

    # create a temporary file
    with tempfile.NamedTemporaryFile(mode="w") as makefile:
        # read the user's file
        with user_file.open(encoding="utf-8") as user_makefile:
            user_data = user_makefile.read()

        # get the path to this package's data to include our "conf.mk"
        include_str = f"include {get_resource('conf.mk')}"

        # if the project name wasn't provided, guess that it's either the name
        # of the parent directory, or that name as a "slug"
        if project_name is None:
            parent = user_file.resolve().parent
            parent_slug = parent.name.replace("-", "_")
            if Path(parent, parent_slug).is_dir():
                project_name = parent_slug
            else:
                project_name = str(parent)

        # build the necessary file data
        data = {
            "PROJ": os.path.basename(project_name),
            "$(PROJ)_DIR": str(directory),
            "MK_AUTO": "1",
        }

        write_data = ""
        for key, item in data.items():
            write_data += f"{key} := {item}" + os.linesep
        write_data += include_str + os.linesep + os.linesep + user_data

        # write the file and return the path
        makefile.write(write_data)
        makefile.flush()
        yield makefile.name


def entry(args: argparse.Namespace) -> int:
    """Execute the requested task."""

    if not args.file.is_file():
        if args.file.name != str(DEFAULT_FILE):
            LOG.error("'%s' not found", args.file)
            return 1
        args.file = get_resource(os.path.join("data", "header.mk"))

    # build the beginning of the invocation args
    invocation_args = ["make", "-C", str(args.dir), "-f"]
    with build_makefile(args.file, args.dir, args.proj) as makefile:
        invocation_args.append(makefile)

        # add each target to the list
        for target in args.targets:
            target_str = target
            if args.prefix and "=" not in target_str:
                target_str = f"{args.prefix}-{target_str}"
            invocation_args.append(target_str)

        # start the process
        LOG.debug(invocation_args)
        try:
            result = subprocess.run(invocation_args, check=True)
            retcode = result.returncode
        except subprocess.CalledProcessError as exc:
            retcode = exc.returncode
        except KeyboardInterrupt:
            retcode = 1

    return retcode


def add_app_args(parser: argparse.ArgumentParser) -> None:
    """Add application-specific arguments to the command-line parser."""

    parser.add_argument("targets", nargs="*", help="targets to execute")
    parser.add_argument(
        "-p", "--prefix", default="", help="a prefix to apply to all targets"
    )
    parser.add_argument(
        "-f",
        "--file",
        default=DEFAULT_FILE,
        type=Path,
        help="file to source user-provided recipes from",
    )
    parser.add_argument(
        "-P", "--proj", help="project name for internal variable use"
    )
